/****************************************************************************
**
** Copyright (C) 2006-2009 fullmetalcoder <fullmetalcoder@hotmail.fr>
**
** This file is part of the Edyuk project <http://edyuk.org>
** 
** This file may be used under the terms of the GNU General Public License
** version 3 as published by the Free Software Foundation and appearing in the
** file GPL.txt included in the packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
****************************************************************************/

#include "qmarksmanager.h"

/*!
	\file qmarksmanager.cpp
	\brief Implementation of QMarksManager
	
	\see QMarksManager
*/

/*!
	\ingroup language
	@{
	
	\class QMarksManager
	\brief A class managing line marks accross all managed editors
	
	QMarksManager provides definitions of line marks (id, priority rules, associated formatting...)
	and several core internal features related to marks management. It can read marks definition
	from a simple XML file whose structure is explained below :
	
	
*/

#include "qeditor.h"
#include "qcodeedit.h"

#include "qdocument.h"
#include "qdocument_p.h"
#include "qdocumentline.h"
#include "qdocumentmark.h"

#include <QFile>
#include <QDebug>
#include <QDataStream>
#include <QDomDocument>
#include <QLinearGradient>

QMarksManager* QMarksManager::m_instance = 0;

QMarksManager* QMarksManager::instance()
{
	if ( !m_instance )
		m_instance = new QMarksManager;
	
	return m_instance;
}

void QMarksManager::destroy()
{
	if ( m_instance )
		delete m_instance;
	
	m_instance = 0;
}

QMarksManager::QMarksManager()
 : QObject(0)
{
	qRegisterMetaType<QMarkType>("QMarkType");
}

QMarksManager::~QMarksManager()
{
	
}

static QColor color(const QString& value)
{
	QColor c;
	
	if ( value.contains('@') )
	{
		c = QColor(value.section('@', 0, 0, QString::SectionSkipEmpty));
		c.setAlpha(value.section('@', 1, 1, QString::SectionSkipEmpty).toUInt(0, 16));
	} else {
		c = QColor(value);
	}
	
	return c;
}

/*
	split 2 level S-expr : (a b) b (a  b)
*/
static QList<QStringList> splitSExpr2(const QString& s)
{
	QList<QStringList> l;
	int idx = 0, last = 0, end = s.count();
	const quint16 *unicode = s.utf16();
	
	while ( idx < end )
	{
		const quint16 c = unicode[idx];
		const bool ws = isspace(c);
		
		if ( ws || c == '(' )
		{
			if ( last != idx )
			{
				l << QStringList(s.mid(last, idx - last));
			}
			
			if ( !ws )
			{
				QStringList sl;
				
				while ( ++idx < end )
				{
					const quint16 sc = unicode[idx];
					const bool sws = isspace(sc);
					
					if ( sws || sc == ')' )
					{
						if ( last != idx )
						{
							sl << s.mid(last, idx - last);
							last = idx;
						}
						
						if ( !sws )
							break;
					}
				}
				
				l << sl;
			}
			
			last = idx + 1;
		}
		
		++idx;
	}
	
	if ( last != idx )
	{
		l << QStringList(s.mid(last));
	}
	
	return l;
}

static QPointF position(const QStringList& l)
{
	QPointF p;
	
	static QHash<QString, QPointF> named;
	
	if ( named.isEmpty() )
	{
		named["topLeft"] = QPointF(0, 0);
		named["bottomLeft"] = QPointF(0, 1);
		named["topRight"] = QPointF(1, 0);
		named["bottomRight"] = QPointF(1, 1);
		named["center"] = QPointF(.5, .5);
		named["midLeft"] = QPointF(0, .5);
		named["midRight"] = QPointF(1, .5);
		named["topMid"] = QPointF(.5, 0);
		named["bottomMid"] = QPointF(.5, 1);
	}
	
	if ( l.count() == 1 )
	{
		p = named.value(l.at(0));
	} else if ( l.count() >= 2 ) {
		p = QPointF(l.at(0).toFloat(), l.at(1).toFloat());
	}
	
	return p;
}

/*!
	\brief Load line marks definition from a file
*/
void QMarksManager::loadMarkTypes(const QString& f)
{
	QFile file(f);
	
	if ( !file.open(QFile::ReadOnly | QFile::Text) )
		return;
	
	// TODO : prefer QXmlStreamReader when building against Qt 4.3.0
	
	QDomDocument doc;
	doc.setContent(&file);
	
	QDomNodeList l = doc.documentElement().childNodes();
	
	for ( int i = 0; i < l.count(); ++i )
	{
		QDomElement e = l.at(i).toElement();
		
		if ( e.isNull() || (e.tagName() != "mark") )
			continue;
		
		QMarkType t;
		QDomNodeList c = e.childNodes();
		
		//qDebug("mark {");
		
		for ( int j = 0; j < c.count(); ++j )
		{
			QDomElement attr = c.at(j).toElement();
			
			if ( attr.isNull() )
				continue;
			
			const QString field = attr.tagName();
			QString value = attr.firstChild().toText().data();
			
			//qDebug("\t%s = %s;", qPrintable(field), qPrintable(value));
			
			const bool flag = (value == "true") || value.toUInt();
			
			if ( field == "id" )
			{
				t.id = value;
			} else if ( field == "user" ) {
				t.user = flag;
			} else if ( field == "focus" ) {
				t.focus = flag;
			} else if ( field == "icon" ) {
				t.icon = QPixmap(value);
			} else if ( field == "color" ) {
				/*
					color value can either be
					
					 * a valid value for QColor::setNamedColor() with one exception though :
					an alpha channel indication (unsupported by QColor::setNamedColor()) can
					be appended using '@' followed by a sequence of hex digits (preferabily 
					two of them...)
					 * a linear gradient  : two points and an arbitrary number of color stops
					along the line defined by the two points. Points coords in [0-1] range,
					relative to rect to be filled. Color stops follow the QGradient convention :
					[0-1] relative pos along the line followed by color (see above for valid
					color def)
					
					examples :
						#ff0c80
						#ff0c80@80
						blue
						blue@10
						[(x y) (x y) (0 blue) (1 white)]
					
				*/
				
				QLinearGradient *lg = new QLinearGradient;
				
				t.gradient = lg;
				t.gradient->setCoordinateMode(QGradient::ObjectBoundingMode);
				
				QGradientStops stops;
				QPointF start = QPointF(.5, 0);
				QPointF stop = QPointF(.5, 1);
				
				if ( value.contains(' ') || value.contains('(') )
				{
					// parse gradient def
					QList<QStringList> def = splitSExpr2(value);
					
					if ( def.count() >= 3 )
					{
						start = position(def.takeFirst());
						stop = position(def.takeFirst());
					}
					
					const qreal increment = 1.0 / (qreal)(def.count() - 1);
					qreal last = -increment;
					
					foreach ( const QStringList& stop, def )
					{
						if ( stop.count() == 1 )
						{
							// shortened
							last += increment;
							stops << qMakePair(last, color(stop.at(0)));
						} else {
							last = stop.at(0).toFloat();
							stops << qMakePair(last, color(stop.at(1)));
						}
					}
				} else {
					// prepare gradient for simple color fill
					
					QColor c = color(value);
					stops << qMakePair(0.0, c);
					stops << qMakePair(1.0, c);
				}
				
				lg->setStart(start);
				lg->setFinalStop(stop);
				lg->setStops(stops);
			} else if ( field == "priority" ) {
				t.priority = value.toUInt();
			} else if ( field == "persistency" ) {
				t.persistency = value.toUInt();
			} else if ( field == "rule" ) {
				t.rules << value;
			}
			
		}
		
		m_markTypes << t;
		
		//qDebug("};");
	}
}

/*!
	\brief int -> string mark type identifier conversion
*/
QString QMarksManager::markTypeId(int id) const
{
	return ((id >= 0) && (id < m_markTypes.count())) ? m_markTypes.at(id).id : QString();
}

/*!
	\brief string -> int mark type identifier conversion
*/
int QMarksManager::markTypeId(const QString& id) const
{
	for ( int idx = 0; idx < m_markTypes.count(); ++idx )
		if ( m_markTypes.at(idx).id == id )
			return idx;
	
	return -1;
}

static const QMarkType& _invalid()
{
	static const QMarkType _invalidMarkTypeInstance;
	
	return _invalidMarkTypeInstance;
}

/*!
	\return The mark type definition associated with a given id
*/
const QMarkType& QMarksManager::markType(int id) const
{
	return ((id >= 0) && (id < m_markTypes.count())) ? m_markTypes.at(id) : _invalid();
}

/*!
	\return the mark type definition associated with a given id
*/
const QMarkType& QMarksManager::markType(const QString& id) const
{
	foreach ( const QMarkType& t, m_markTypes )
		if ( t.id == id )
			return t;
	
	return _invalid();
}

/*!
	\return A list of available mark types
	\param context context filter (no filtering is performed if empty)
*/
QStringList QMarksManager::availableMarkTypes(const QString& context) const
{
	QStringList l;
	
	foreach ( const QMarkType& t, m_markTypes )
	{
		if (
				context.count()
			&&
				(
					!t.user
				||
					(
						t.rules.contains("#out")
					&&
						!t.rules.contains(context)
					)
				||
					(
						t.rules.contains("#in")
					&&
						t.rules.contains("!" + context)
					)
				)
			)
		{
			//qDebug("mark[%s] mismatched", qPrintable(t.id));
		} else {
			l << t.id;
		}
	}
	
	return l;
}

/*!
	\return the mark that has the highest priority among a list of marks
*/
int QMarksManager::priority(const QList<int>& marks) const
{
	int highest = -1;
	int mark = marks.isEmpty() ? -1 : marks.at(0);
	
	for ( int i = 0; i < m_markTypes.count(); ++i )
	{
		if ( marks.contains(i) && (m_markTypes.at(i).priority > highest) )
		{
			mark = i;
			highest = m_markTypes.at(i).priority;
		}
	}
	
	return mark;
}

/*!
	\return the mark that has the highest priority among a list of marks
*/
QString QMarksManager::priority(const QStringList& marks) const
{
	QString mark;
	int highest = -1;
	
	foreach ( const QMarkType& t, m_markTypes )
	{
		if ( marks.contains(t.id) && (t.priority > highest) )
		{
			mark = t.id;
			highest = t.priority;
		}
	}
	
	return (mark.count() || !marks.count()) ? mark : marks.at(0);
}

/*!
	\brief Useless for now
*/
QList<QStringList> QMarksManager::marksLayout(const QString& context) const
{
	QList<QStringList> l;
	const QStringList mt = availableMarkTypes(context);
	
	foreach ( QString id, mt )
	{
		l << QStringList(id);
	}
	
	/*
	foreach ( QLineMarkType t, availableMarks(context) )
	{
		
	}
	*/
	
	return l;
}

static bool inScope(QDocumentCursorHandle *h, const QDocumentCursor& c)
{
	return h->hasSelection() ? h->intersect(c).isValid() : h->lineNumber() == c.lineNumber();
}

/*!
	\internal
*/
void QMarksManager::cursorMoved(QEditor *e)
{
	QDocumentPrivate *d = e->document()->impl();
	QList<QDocumentMark*>& marks = d->m_marks;
	
	int idx = 0;
	
	while ( idx < marks.count() )
	{
		QDocumentMark *m = marks.at(idx);
		const QMarkType& t = markType(m->type());
		
		if ( !t.persistency || (t.persistency == 1 && !inScope(m->m_location, e->cursor())) )
		{
			QList<QDocumentLineHandle*> preLines = m->m_location->selectionSpan();
			
			foreach ( QDocumentLineHandle *h, preLines )
			{
				QList<QDocumentMark*>& ml = d->m_marksLookup[h];
				
				ml.removeAll(m);
				
				if ( ml.isEmpty() )
					d->m_marksLookup.remove(h);
				
			}
			
			marks.removeAt(idx);
		} else {
			++idx;
		}
	}
	
	#if 0
	foreach ( const QDocumentMark *m, m_lineMarks )
	{
		QLineMarkType t = markType(lmh.mark);
		
		if (
				(e->fileName() != lmh.file)
		//	||
		//		(e->document() != lmh.line->document())
			||
				(t.persistency == 2)
			)
			continue;
		
		#warning FIXME
		/*
		if ( !t.persistency || (lmh.line != e->cursor().line().handle()) )
		{
			removeLineMark(lmh);
			cursorMoved(e);
			break;
		}
		*/
	}
	#endif
}

/*! @} */
